/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"
#include "mutt_regex.h"

/* HP-UX and ConvexOS don't have this macro */
#ifndef S_ISLNK
#define S_ISLNK(x) (((x) & S_IFMT) == S_IFLNK ? 1 : 0)
#endif

#define M_FOLDERFMT "---Mutt: File Browser [Directory: %s]"

struct folder_file {
  mode_t mode;
  char *name;
  char *desc;
};

/*
 * Global variables for this package.
 */
static struct folder_file *Entries = NULL;
static short LastEntry = 0; /* number of real entries */
static short MaxEntry = 0;  /* max entry */
static char LastDir[_POSIX_PATH_MAX] = "";

/*
 * Frees up the memory allocated for the local-global variables.
 */
static void destroy_global (void)
{
  int c;

  for (c = 0; c < LastEntry; c++)
  {
    safe_free ((void **)&Entries[c].name);
    safe_free ((void **)&Entries[c].desc);
  }
  safe_free ((void **)&Entries);
}

int compare_entries (const void *a, const void *b)
{
  struct folder_file *pa = (struct folder_file *)a;
  struct folder_file *pb = (struct folder_file *)b;

  return (strcmp (pa->name, pb->name));
}

static int examine_directory (const char *d)
{
  struct stat s;
  DIR *dp;
  struct dirent *de;
  struct passwd *pw;
  struct group *gr;
  char buffer[LONG_STRING];
  char date[16];
  char *fmt;
  time_t tnow;

  if (stat (d, &s) == -1)
  {
    mutt_perror (d);
    return (-1);
  }

  if (!S_ISDIR (s.st_mode))
  {
    beep ();
    mutt_error ("Warning: %s is not a directory", buffer);
    return (-1);
  }

  LastEntry = 0;
  MaxEntry = 256;
  Entries = (struct folder_file *)safe_malloc (sizeof (struct folder_file) * MaxEntry);

  if ((dp = opendir (d)) == NULL) return (-1);

  while ((de = readdir (dp)) != NULL)
  {
    if (strcmp (de->d_name, ".") == 0) continue;    /* we don't need . */
    
    sprintf (buffer, "%s/%s", d, de->d_name);
    if (lstat (buffer, &s) == -1) continue;
    
    if ((! S_ISREG (s.st_mode)) && (! S_ISDIR (s.st_mode))) continue;
    
    pw = getpwuid (s.st_uid);
    gr = getgrgid (s.st_gid);

    tnow = time (NULL);
    fmt = tnow - s.st_mtime < 31536000 ? "%b %d %H:%M" : "%b %d  %Y";
    strftime (date, sizeof (date), fmt, localtime (&s.st_mtime));

    snprintf (buffer, sizeof (buffer), "%c%c%c%c%c%c%c%c%c%c %2d %-8.8s %-8.8s %8ld %s %s%s",
	      S_ISDIR(s.st_mode) ? 'd' : (S_ISLNK(s.st_mode) ? 'l' : '-'),

	      (s.st_mode & S_IRUSR) != 0 ? 'r': '-',
	      (s.st_mode & S_IWUSR) != 0 ? 'w' : '-',
	      (s.st_mode & S_ISUID) != 0 ? 's' : (s.st_mode & S_IXUSR) != 0 ? 'x': '-',

	      (s.st_mode & S_IRGRP) != 0 ? 'r' : '-',
	      (s.st_mode & S_IWGRP) != 0 ? 'w' : '-',
	      (s.st_mode & S_ISGID) != 0 ? 's' : (s.st_mode & S_IXGRP) != 0 ? 'x': '-',

	      (s.st_mode & S_IROTH) != 0 ? 'r' : '-',
	      (s.st_mode & S_IWOTH) != 0 ? 'w' : '-',
	      (s.st_mode & S_ISVTX) != 0 ? 't' : (s.st_mode & S_IXOTH) != 0 ? 'x': '-',
	      s.st_nlink,
	      pw ? pw->pw_name : "(null)",
	      gr ? gr->gr_name : "(null)",
	      (long)s.st_size,
	      date,
	      de->d_name,
	      S_ISDIR (s.st_mode) ? "/" : ((s.st_mode & S_IXUSR) != 0 ? "*" : ""));


    if (LastEntry == MaxEntry)
    {
      /* need to allocate more space */
      safe_realloc ((void **)&Entries, sizeof (struct folder_file) * (MaxEntry += 256));
    }

    Entries[LastEntry].mode = s.st_mode;
    Entries[LastEntry].name = safe_strdup (de->d_name);
    Entries[LastEntry].desc = safe_strdup (buffer);
    
    LastEntry++;
  }
  closedir (dp);  
  qsort (Entries, LastEntry, sizeof (struct folder_file), compare_entries);
  return 0;
}

int select_file_search (const char *pat, MUTTMENU *menu)
{
  int i;
  regex_t re;
  
  if (REGCOMP (re, pat, 0) != 0)
  {
    beep ();
    mutt_error ("Error in regexp!");
    return (-1);
  }
  for (i = menu->current + 1 ; i < menu->max ; i++)
  {
    if (REGEXEC (re, Entries[i].name) == 0)
    {
      REGFREE (re);
      return (i);
    }
  }
  beep ();
  mutt_error ("Not found.");
  REGFREE (re);
  return (-1);
}

void folder_makeEntry (char *s, int num, size_t slen)
{
  snprintf (s, slen, "%-*.*s", COLS, COLS, Entries[num].desc);
}

static void init_menu (MUTTMENU *menu)
{
  char buffer[STRING];

  menu->current = 0;
  menu->top = 0;
  menu->max = LastEntry;
  safe_free ((void **)&menu->title);
  snprintf (buffer, sizeof (buffer), M_FOLDERFMT, LastDir);
  menu->title = safe_strdup (buffer);
  menu->redraw = REDRAW_FULL;
}

void mutt_select_file (char *f, size_t flen)
{
  char buffer[STRING];
  MUTTMENU *menu;
  struct stat st;
  
  if (!LastDir[0])
  {
    strfcpy (LastDir, Maildir, sizeof (LastDir));
    mutt_expand_path (LastDir, sizeof (LastDir));
  }

  *f = 0;

  if (examine_directory (LastDir) == -1) return;

  menu = mutt_newMenu ("");
  menu->menu = MENU_FOLDER;
  menu->makeEntry = folder_makeEntry;
  menu->search = select_file_search;
  init_menu (menu);

  FOREVER
  {
    switch (mutt_menuLoop (menu))
    {
      case OP_GENERIC_SELECT_ENTRY:

	if (S_ISDIR (Entries[menu->current].mode))
	{
	  if (strcmp (Entries[menu->current].name, "..") == 0)
	  {
	    if (strcmp ("..", LastDir + strlen (LastDir) - 2) == 0)
	      strcat (LastDir, "/..");
	    else
	    {
	      char *p = strrchr (LastDir + 1, '/');

	      if (p)
		*p = 0;
	      else
	      {
		if (LastDir[0] == '/')
		  LastDir[1] = 0;
		else
		  strcat (LastDir, "/..");
	      }
	    }
	  }
	  else
	    sprintf (LastDir + strlen (LastDir), "/%s", Entries[menu->current].name);

	  destroy_global ();
	  if (examine_directory (LastDir) == -1) return;
	  init_menu (menu);
	  break;
	}
	snprintf (f, flen, "%s/%s", LastDir, Entries[menu->current].name);

	/*
	 * Fall through to OP_GENERIC_EXIT
	 */

      case OP_GENERIC_EXIT:

	destroy_global ();
	mutt_menuDestroy (&menu);
	return;

      case OP_CHANGE_DIRECTORY:

	strfcpy (buffer, LastDir, sizeof (buffer));
	if (ci_get_field ("Chdir to: ", buffer, sizeof (buffer), M_FILE) == 0 &&
	    buffer[0])
	{
	  mutt_expand_path (buffer, sizeof (buffer));
	  if (stat (buffer, &st) == 0)
	  {
	    if (S_ISDIR (st.st_mode))
	    {
	      strfcpy (LastDir, buffer, sizeof (LastDir));
	      destroy_global ();
	      if (examine_directory (LastDir) == 0)
		init_menu (menu);
	      else
	      {
		beep ();
		mutt_error ("Error scanning directory.");
		destroy_global ();
		mutt_menuDestroy (&menu);
		return;
	      }
	    }
	    else
	    {
	      beep ();
	      mutt_error ("%s is not a directory.", buffer);
	    }
	  }
	  else
	  {
	    beep ();
	    mutt_perror (buffer);
	  }
	}
	break;
    }
  }
  /* not reached */
}
