/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"
#include "rfc1524.h"
#include "rfc_gen.h" /* pms@romance.kaist.ac.kr */

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>

struct attachptr {
  BODY *content;
  struct attachptr *next;
};

struct attachptr **idx = NULL;

static struct attachptr *gen_attach_list (BODY *m)
{
  struct attachptr *top = NULL, *end = NULL, *new;

  while (m)
  {
    if (m->type == TYPEMULTIPART)
    {
      if (top)
        end->next = gen_attach_list(m->parts);
      else
        end = top = gen_attach_list(m->parts);
      while (end->next) end = end->next;
    }
    else
    {
      new = (struct attachptr *)safe_malloc(sizeof(struct attachptr));
      new->next = NULL;
      new->content = m;

      if (top)
      {
        end->next = new;
        end = new;
      }
      else
        top = end = new;

      if (m->type == TYPEMESSAGE && !strcasecmp(m->subtype, "rfc822") &&
	  is_multipart (m->parts))
      {
	end->next = gen_attach_list(m->parts);
	while (end->next) end = end->next;
      }
    }
    m = m->next;
  }
  return (top);
}

static int can_print (BODY *m)
{
  if (m->type == TYPETEXT)
  {
    if (strcasecmp ("plain", m->subtype) == 0) return 1;
  }
  else if (m->type == TYPEAPPLICATION)
  {
    if (strcasecmp ("postscript", m->subtype) == 0) return 1;
  }
  return 0;
}

/*
** Support nation specific RFCs - pms@romance.kaist.ac.kr
*/
static void mutt_decode_attachment (BODY *b, STATE *s)
{
  fseek (s->fpin, b->offset, 0);
  switch (b->encoding)
  {
    case ENCQUOTEDPRINTABLE:
      mutt_decode_quoted (s, b->length, mutt_is_text_type (b->type, b->subtype));
      break;
    case ENCBASE64:
      mutt_decode_base64 (s, b->length, mutt_is_text_type (b->type, b->subtype));
      break;
    case ENC8BIT:
    case ENC7BIT:
      if(mutt_decode_own_rfc(s, b->length, 
         mutt_get_parameter("charset", b->parameter))) break;
    default:
      mutt_decode_xbit (s, b->length);
      break;
  }
}

static int save_attachment (BODY *m, char *path)
{
  STATE s;

  if (access (path, F_OK) == 0)
  {
    mutt_error ("File already exists!");
    return (-1);
  }
  memset (&s, 0, sizeof (s));
  if ((s.fpout = fopen (path, "w")) == NULL)
  {
    mutt_perror ("fopen");
    return (-1);
  }
  fseek ((s.fpin = Context->folder->fp), m->offset, 0);
  mutt_decode_attachment (m, &s);
  if (fclose (s.fpout) == 0) return 0;
  dprint(1,(debugfile, "save_attachment: fclose() returned non-zero! (errno=%d)\n", errno));
  mutt_perror ("fclose");                                                      
  return (-1);
}

static void pipe_attachment (BODY *b, const char *path)
{
  STATE s;

  memset (&s, 0, sizeof (STATE));
  s.fpin = Context->folder->fp;
  endwin ();
  mutt_create_filter (path, &s.fpout, NULL);
  mutt_decode_attachment (b, &s);
  fclose (s.fpout);
  wait (NULL);
  ci_any_key_to_continue (NULL);
}

void mutt_display_attachment (BODY *a, int type)
{
  STATE *s;
  char tempfile[_POSIX_PATH_MAX] = "";
  char descrip[SHORT_STRING];

  if (a->filename) strfcpy (tempfile, a->filename, sizeof (tempfile));
  mutt_adv_mktemp (tempfile);

  if (type == M_MAILCAP || (type == M_REGULAR && mutt_needs_mailcap (a)))
  {
    char type[STRING];
    char command[STRING];
    rfc1524_entry *entry = rfc1524_new_entry();

    snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
    if (rfc1524_mailcap_lookup (a,type, entry))
    {
      if (entry->command)
      {
	strfcpy(command,entry->command,sizeof(command));
	if (!rfc1524_expand_command(a,tempfile, type, command, sizeof(command)))
	{
	  /* rfc1524_expand_command returns 0 if the file is required */
	  save_attachment (a, tempfile);
	  if (entry->copiousoutput)
	  {
	    /* Use a pager: The assumption is the info is coming from 
	     * stdout of the command without a pager, so we use our own
	     */
	    char buffer[LONG_STRING];
	    char outfile[_POSIX_PATH_MAX] = "";
	    FILE *fpin;
	    FILE *tfp;
	    int len;

	    mutt_mktemp(outfile);
	    tfp = fopen(outfile,"w");
            mutt_create_filter (command, NULL, &fpin);
            endwin ();
	    while ((len = fread (buffer, 1, sizeof (buffer), fpin)) > 0)
	      fwrite (buffer, 1, len, tfp);
            wait (NULL);
            fclose (fpin);
	    fclose (tfp);

            if (a->description)
	      snprintf (descrip, sizeof(descrip),
			"---Command: %-20.20s Description: %s",command,a->description);
	    else
	      snprintf (descrip, sizeof (descrip),
			"---Command: %-30.30s Attachment: %s/%s", command, 
			TYPE (a->type), a->subtype);
	    mutt_do_pager (descrip, outfile);
	  }
	  else
	  {
	    endwin();
	    mutt_system (command);
	  }
	}
	else
	{
	  /* rfc1524_expand_command returns 1 if it takes a pipe */
	  pipe_attachment (a, command);
	}
      }
    }
    else
    {
      /* no mailcap type defined */
    }
    mutt_unlink (tempfile);
    return;
  }
  else
  {
    s = mutt_new_state ();
    s->fpout = fopen (tempfile, "w");
    s->fpin = Context->folder->fp;

    if (type == M_AS_TEXT)
    {
      /* don't try to do any fancy handling, just let me see the raw data! */
      mutt_decode_attachment (a, s);
    }
    else
    {
      s->displaying = 1;
      mutt_body_handler (a, s);
    }
    fclose (s->fpout);
    mutt_free_state (&s);

     if (a->description)
       strfcpy (descrip, a->description, sizeof (descrip));
     else
       snprintf (descrip, sizeof (descrip),
		 "---Attachment: %s/%s", TYPE (a->type), a->subtype);
     mutt_do_pager (descrip, tempfile);
   }
}

void attach_makeEntry (char *b, int num, size_t blen)
{
  char t[SHORT_STRING];
  BODY *m = idx[num]->content;

  snprintf (t, sizeof(t), "%s/%s", TYPE(m->type), m->subtype);
  snprintf (b, blen, " %2d %-20.20s  %-6.6s (%6ld) %s", num+1, t,
	    ENCODING(m->encoding), m->length,
	    m->description ? m->description : (m->filename ? m->filename : ""));
}

void mutt_view_attachments (HEADER *hdr)
{
  struct attachptr *list, *plist;
  int idxlen = 0, count = 0;
  char buffer[STRING];
  MUTTMENU *menu;
  BODY *cur = hdr->content;
  
  /* make sure we have parsed this message */
  mutt_parse_mime_message (hdr);

  plist = list = gen_attach_list (cur);
  idx = safe_malloc (sizeof (struct attachptr *) * ((idxlen = 5)));
  idxlen = 5;

  for (count=0; plist; count++, plist = plist->next)
  {
    if (count == idxlen)
      safe_realloc ((void **)&idx, ((idxlen += 5)) * sizeof(struct attachptr *));
    idx[count] = plist;
  }

  snprintf (buffer, sizeof (buffer), "---Mutt: Attachments [%d part%s]", count, count != 1 ? "s" : "");
  menu = mutt_newMenu (buffer);
  menu->max = count;
  menu->makeEntry = attach_makeEntry;
  menu->menu = MENU_ATTACH;

  FOREVER
  {
    switch (mutt_menuLoop (menu))
    {
      case OP_GENERIC_SELECT_ENTRY:

	mutt_display_attachment (idx[menu->current]->content, M_REGULAR);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_ATTACH_VIEW_MAILCAP:

	mutt_display_attachment (idx[menu->current]->content,  M_MAILCAP);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_ATTACH_VIEW_TEXT:

	mutt_display_attachment (idx[menu->current]->content,  M_AS_TEXT);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_PRINT:

	if (can_print (idx[menu->current]->content))
	  pipe_attachment (idx[menu->current]->content, Print);
	else
	{
	  beep ();
	  mutt_error ("I don't know how to print that!");
	}
	break;

      case OP_ATTACH_PIPE_ENTRY:

	buffer[0] = 0;
	if (ci_get_field ("Pipe to: ", buffer, sizeof (buffer), 0) == 0 &&
	    buffer[0])
	{
	  mutt_expand_path (buffer, sizeof (buffer));
	  pipe_attachment (idx[menu->current]->content, buffer);
	  ci_any_key_to_continue (NULL);
	}
	break;

      case OP_ATTACH_SAVE_ENTRY:

	if (idx[menu->current]->content->filename)
	  strfcpy(buffer, idx[menu->current]->content->filename, sizeof(buffer));
	else
	  buffer[0] = 0;
	if (ci_get_field ("Save to file: ", buffer, sizeof (buffer), M_FILE) != 0 ||
	    !buffer[0])
	  break;
	mutt_expand_path (buffer, sizeof (buffer));
	if (save_attachment (idx[menu->current]->content, buffer) == 0)
	  mutt_error ("Attachment saved.");
	else
	{
	  beep ();
	  mutt_error ("Error while saving attachment!");
	}
	break;

      case OP_GENERIC_EXIT:

	while (count-- > 0) safe_free((void **)&idx[count]);
	safe_free ((void **)&idx);
	mutt_menuDestroy (&menu);
	return;
    }
  }
  /* not reached */
}
